/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/

package ca.mcgill.cs.stg.fraco;

import ca.mcgill.cs.stg.fraco.Properties.IdentifierPropertiesResolver;
import ca.mcgill.cs.stg.fraco.parsers.FieldParser;
import ca.mcgill.cs.stg.fraco.parsers.LocalsParser;
import ca.mcgill.cs.stg.fraco.parsers.MethodParser;
import ca.mcgill.cs.stg.fraco.parsers.ParseAstNodes;
import ca.mcgill.cs.stg.fraco.parsers.TypeParser;
import ca.mcgill.cs.stg.fraco.utils.CommentType;
import ca.mcgill.cs.stg.fraco.utils.Comments;
import ca.mcgill.cs.stg.fraco.utils.Identifier;

import ca.mcgill.cs.stg.fraco.utils.Scope;
import ca.mcgill.cs.stg.fraco.utils.Type;
import ca.mcgill.cs.stg.fraco.visitors.ClassVisitor;
import ca.mcgill.cs.stg.fraco.visitors.CommentVisitor;
import ca.mcgill.cs.stg.fraco.visitors.FieldVisitor;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;

import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;

import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.*;

import java.util.ArrayList;

import java.util.List;
import java.util.Optional;

import java.util.concurrent.Executors;

/**
 * Builds the AST of a project present in the workspace of plugin. Parses all
 * the required elements and gathers information required for further
 * processing. 1. stores all the comments present in the project in a list. 2.
 * stores all the identifiers present in the project in a list.
 * 
 * @author inderjot
 *
 */
public class AstBuilder {

	/**
	 * Collection of objects of class Comments to store all types of comments
	 * present in a project.
	 */
	public static List<Comments> _commentsTable;

	/**
	 * needed for experiment set up. Holds all the Identifier present in a
	 * project.
	 */
	public static List<Identifier> _identifiersTable;

	/** represents the identifier renamed by the developer */
	public static IJavaElement _renamedElement;

	public static String _newName;

	/**
	 * stores the string value of compilation unit under processing. Source
	 * value is needed to retrieve LINE and BLOCK comments text.
	 */
	private static String _source;

	/**
	 * list of plain comments' text, needed for further processing of comments.
	 */
	public static List<String> _commentList;

	/*
	 * instance of class Identifier used to capture all the required properties
	 * of a renamed java element.
	 */
	public static Identifier _identifier;

	/**
	 * Holds the value of compilation Unit under process at a particular
	 * instance of time.
	 */
	public static CompilationUnit _currentCompilationUnit;

	/**
	 * Holds the value of ICompilationUnit under process at a particular
	 * instance of time.
	 */
	public static ICompilationUnit _cunit;

	/**
	 * holds the name of project package under processing.
	 */
	public static String _currentPackage;

	/**
	 * the name of project under processing.
	 */
	public static String _projectName;

	// counter for the number of identifiers
	public static int numberOfIdentifiers = 0;

	/**
	 * Entry point to start the main function of plug in. Initializes the
	 * identifiersTable and comment list on every call.
	 */
	public AstBuilder() {
		_identifiersTable = new ArrayList<Identifier>();
		_commentList = new ArrayList<String>();
		_commentsTable = new ArrayList<Comments>();
		_renamedElement = null;
		_identifier = null;
	}

	/**
	 * Initializes the main functionality of plugin. Starts AST Parsing and data
	 * collection.
	 */
	public void initializePluginProcessing() {
		// collects information about the renamed identifier.
		IdentifierPropertiesResolver idp = new IdentifierPropertiesResolver();
		_identifier = idp.setCurrentIdentifier();

		// initiates parsing AST of of the project related to renamed
		// identifier, present in the workspace.
		initASTParsing();

		// if renamed element is not null, continue fragility detection after
		// AST parsing.
		if (_renamedElement != null) {
			ParseAstNodes.markAllInnerCommentsAsLocal(_identifier);
			DetectFragileComments.InitFragilityDetection();
			_identifier = null;
			_commentsTable.clear();
		}
	}

	/**
	 * Getting workspace and deleting all the existing markers if any.
	 */
	private synchronized void initASTParsing() {
		IWorkspace workspace = ResourcesPlugin.getWorkspace();
		IWorkspaceRoot root = workspace.getRoot();

		// deleting existing markers before creating new ones.
		// int depth = IResource.DEPTH_INFINITE;
		/*
		 * try {
		 * 
		 * root.deleteMarkers(IMarker.PROBLEM, true, depth);
		 * 
		 * } catch (CoreException e1) { // TODO Auto-generated catch block
		 * e1.printStackTrace(); }
		 */

		// Get all projects in the workspace
		IProject[] projects = root.getProjects();
		// Loop over all projects
		for (IProject project : projects) {
			try {
				// _commentsTable= new ArrayList<Comments>();
				// only parse the project which contains the identifier
				if (_renamedElement != null
						&& project.getName().equals(_renamedElement.getJavaProject().getElementName())) {
					getProjectInfo(project);
					break;
				} else if (_renamedElement == null) {
					getProjectInfo(project);
				}
			} catch (CoreException e) {
				e.printStackTrace();
			}
		}

	}

	private void getProjectInfo(IProject project) throws CoreException, JavaModelException {
		// System.out.println("Working in project " + project.getName());
		// check if we have a Java project
		_projectName = project.getName();
		if (project.isNatureEnabled("org.eclipse.jdt.core.javanature")) {
			IJavaProject javaProject = JavaCore.create(project);
			getPackagesInfo(javaProject);
		}
	}

	private void getPackagesInfo(IJavaProject javaProject) throws JavaModelException {
		IPackageFragment[] packages = javaProject.getPackageFragments();
		for (IPackageFragment mypackage : packages) {
			// Package fragments include all packages in the
			// classpath
			// We will only look at the package from the source
			// folder
			// K_BINARY would include also included JARS, e.g.
			// rt.jar
			if (mypackage.getPath().toString().toLowerCase().contains("test")
					|| mypackage.getPath().toString().toLowerCase().contains("resources"))
				continue;
			if (mypackage.getKind() == IPackageFragmentRoot.K_SOURCE) {
				// System.out.println("Parsing Package " +
				// mypackage.getElementName());
				getCompilationUnitList(mypackage);
			}
		}
	}

	/*
	 * Checking if the package is non-test package, and if the files traversed
	 * are non-test files or not.
	 */
	private void getCompilationUnitList(IPackageFragment mypackage) throws JavaModelException {
		for (ICompilationUnit unit : mypackage.getCompilationUnits()) {
			_currentPackage = mypackage.getElementName();
			_cunit = unit;
			// create AST for the compilation Unit.
			createAST(unit, mypackage.getElementName());
		}
	}

	public static CompilationUnit parse(ICompilationUnit unit) {
		ASTParser parser = ASTParser.newParser(AST.JLS8);
		parser.setKind(ASTParser.K_COMPILATION_UNIT);
		parser.setSource(unit);
		parser.setResolveBindings(true);
		final CompilationUnit cu = (CompilationUnit) parser.createAST(null);

		return cu;
	}

	/**
	 * Reads a ICompilationUnit and creates the AST DOM for manipulating the
	 * Java source file
	 * 
	 * @param unit
	 *            is the compilation unit needed to create AST.
	 * @param packageName
	 * @throws JavaModelException
	 */
	private void createAST(ICompilationUnit unit, String packageName) throws JavaModelException {

		// now create the AST for the ICompilationUnits
		CompilationUnit compilationUnit = parse(unit);
		// used to get line and block comments text
		_currentCompilationUnit = compilationUnit;
		_source = unit.getSource();
		_commentList = new ArrayList<String>();

		// parse comments present in every package
		parseComments(packageName);

		// only parsing the package which contains renamed identifier. Need
		// parent class name ,modifiers etc. to determine scope.
		if (_renamedElement != null && _identifier.get_packageName().equals(packageName)) {
			int changedElementType = _renamedElement.getElementType();
			switch (changedElementType) {
			case IJavaElement.METHOD:
				parseMethods();
				break;
			case IJavaElement.FIELD:
				parseFields();
				break;
			case IJavaElement.LOCAL_VARIABLE:
				parseLocalVariables();
				break;
			case IJavaElement.CLASS_FILE:
				parseClasses();
				break;
			case IJavaElement.TYPE:
				parseTypes();
			default:
				break;
			}
		}
	}

	private synchronized void parseLocalVariables() {
		LocalsParser localParser = new LocalsParser();
		localParser.parseLocalVariables();
	}

	private synchronized void parseMethods() {
		MethodParser methodParser = new MethodParser();
		methodParser.parseMethods(Optional.empty());
	}

	private synchronized void parseTypes() {
		TypeParser typeParser = new TypeParser();
		typeParser.parseTypes();
	}

	/**
	 * Parse all field declarations, enum constants and field access type AST
	 * nodes in the current compilation Unit.
	 * 
	 * @param compilationUnit
	 */
	private synchronized void parseFields() {
		FieldVisitor visitor = new FieldVisitor();
		_currentCompilationUnit.accept(visitor);
		FieldParser fieldParser = new FieldParser();
		fieldParser.parseFieldDeclarations(visitor);
		fieldParser.parseEnumConstants(visitor);
		// fieldParser.parseFieldAccess(visitor);
	}

	/**
	 * Parse all class instance creations and anonymous type declarations AST
	 * nodes in compilation Unit.
	 * 
	 * @param compilationUnit
	 */
	private synchronized void parseClasses() {
		ClassVisitor visitor = new ClassVisitor();
		_currentCompilationUnit.accept(visitor);
		/*
		 * for (ClassInstanceCreation classes: visitor.getClassInstances()) {
		 * String className=classes.getClass().getName(); int
		 * mod=classes.getClass().getModifiers(); int
		 * index=_currentCompilationUnit.firstLeadingCommentIndex(classes);
		 * 
		 * String parentName=getClassName(classes.getParent()); int
		 * classModifier=getClassModifier(parentName,classes.getParent());
		 * if(_renamedElement==null) {
		 * addIdenitifierToList(className,mod,classModifier,parentName,Scope.
		 * CLASS,null); addProximityComment(index,className); } else
		 * if(_renamedElement!=null &&
		 * className.equals(_renamedElement.getElementName())) {
		 * _identifier.set_type(Scope.CLASS);
		 * _identifier.set_modifier(mapModifier(mod));
		 * _identifier.set_parentModifier(mapModifier(classModifier));
		 * //_identifier.set_cuName(_currentCompilationUnit.getJavaElement().
		 * getElementName());
		 * 
		 * traverseLeadingComments(index); break;
		 * 
		 * } }
		 */

		for (AnonymousClassDeclaration classes : visitor.getAnonymousClasses()) {
			String className = classes.getClass().getName();
			int index = _currentCompilationUnit.firstLeadingCommentIndex(classes);
			int mod = classes.getClass().getModifiers();

			String parentName = ParseAstNodes.getClassName(classes.getParent());
			int classModifier = ParseAstNodes.getClassModifier(parentName, classes.getParent());
			if (_renamedElement != null && className.equals(_renamedElement.getElementName())) {

				_identifier.set_type(Type.CLASS);
				_identifier.set_modifier(ParseAstNodes.mapModifier(mod));
				_identifier.set_parentModifier(ParseAstNodes.mapModifier(classModifier));
				ParseAstNodes.traverseLeadingComments(index);
				break;
			}
		}
	}

	/**
	 * Parse all comment type AST nodes in compilation Unit and store them in a
	 * list of type Comments .
	 * 
	 * @param compilationUnit
	 */
	@SuppressWarnings("unchecked")
	private synchronized void parseComments(String packageName) {
		for (Comment comment : (List<Comment>) _currentCompilationUnit.getCommentList()) {
			Comments newComment = new Comments();
			if (comment.isDocComment()) {
				// _commentList.add(comment.toString());
				CommentVisitor visitor = new CommentVisitor(_currentCompilationUnit, _source, _commentList);
				comment.accept(visitor);
				newComment.set_tags(visitor.getTags());
				newComment.set_comment(comment.toString());
				_commentList.add(newComment.get_comment());
				newComment.set_uid((newComment.get_comment()).hashCode());
				newComment.set_path(_currentCompilationUnit.getJavaElement().getPath().toString());

				if (comment.toString().contains("@")) {
					newComment.set_type(CommentType.JDOCTAG);
				} else {
					newComment.set_type(CommentType.JDOCFREE);
				}
			} else if (comment.isBlockComment()) {

				comment.accept(new CommentVisitor(_currentCompilationUnit, _source, _commentList));
				newComment.set_comment(_commentList.get(_commentList.size() - 1));
				newComment.set_tags(null);
				newComment.set_uid((newComment.get_comment()).hashCode());
				newComment.set_path(_currentCompilationUnit.getJavaElement().getPath().toString());
				newComment.set_type(CommentType.BLOCK);
			} else {
				comment.accept(new CommentVisitor(_currentCompilationUnit, _source, _commentList));
				newComment.set_comment(_commentList.get(_commentList.size() - 1));
				newComment.set_uid((newComment.get_comment()).hashCode());
				newComment.set_tags(null);
				newComment.set_path(_currentCompilationUnit.getJavaElement().getPath().toString());
				newComment.set_type(CommentType.LINE);
			}

			setCommentProperties(newComment, packageName, comment);
			if (!isDuplicate(newComment))
				_commentsTable.add(newComment);

		}
	}

	private void setCommentProperties(Comments newComment, String packageName, Comment comment) {

		newComment.set_lineNumber(_currentCompilationUnit.getLineNumber(comment.getStartPosition()));
		newComment.set_packageName(packageName);
		newComment.set_cUnit(_cunit);
		newComment.set_cuName(_currentCompilationUnit.getJavaElement().getElementName());

		ASTNode node = comment.getParent();
		if (node != null && !comment.isLineComment()) {
			if (node instanceof MethodDeclaration) {
				newComment.set_methodName(((MethodDeclaration) node).getName().getIdentifier());
				node = node.getParent();
			}
			if (ParseAstNodes.getClassName(node) != null) {
				newComment.set_className(ParseAstNodes.getClassName(node, true));
				node = node.getParent();
				newComment.set_parentClassName(ParseAstNodes.getClassName(node, true));
			} else {
				node = node.getParent();
				newComment.set_className(ParseAstNodes.getClassName(node, true));
				node = node.getParent();
				newComment.set_parentClassName(ParseAstNodes.getClassName(node, true));
			}

		}
	}

	private boolean isDuplicate(Comments newComment) {
		for (Comments comment : _commentsTable) {
			if (comment.get_comment().equals(newComment.get_comment())
					&& comment.get_cuName().equals(newComment.get_cuName())
					&& comment.get_lineNumber() == newComment.get_lineNumber()) {
				return true;
			}
		}
		return false;
	}
}
