/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/
package ca.mcgill.cs.stg.fraco.Properties;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;

import ca.mcgill.cs.stg.fraco.AstBuilder;
import ca.mcgill.cs.stg.fraco.utils.CommentType;
import ca.mcgill.cs.stg.fraco.utils.Comments;
import ca.mcgill.cs.stg.fraco.utils.Identifier;
import ca.mcgill.cs.stg.fraco.utils.Match;
import ca.mcgill.cs.stg.fraco.utils.Scope;

/**
 * This class filters the comments based on the scope of an identifier. And also
 * publishes the results on user interface by creating markers to be shown in
 * problems view of eclipse.
 * 
 * @author inder
 *
 */
public class ResultsFiltering {

	List<Comments> _comments;
	Identifier _identifier;
	private int _lineNumber;
	private IResource _resource;
	private List<Comments> _idCommentList;
	private Match _currentMatch;

	public ResultsFiltering() {
	}

	/**
	 * Applies scoping ca.mcgill.cs.stg.fraco.rules on comments based on the
	 * scope and type of identifier.
	 * 
	 * @param comments
	 *            the list of all comments in the source code of the project.
	 * @param identifier
	 *            the identifier under processing.
	 * @return list of comments in scope of the identifier.
	 */
	public synchronized void ApplyCommentScopeRules(List<Comments> comments, Identifier identifier) {
		this._comments = comments;
		this._identifier = identifier;
		this._idCommentList = new ArrayList<Comments>();
		if (_identifier.get_scope() == Scope.GLOBAL) {
			_identifier.set_comments(comments);
		} else {
			for (Comments comment : _comments) {
				FilterComment(comment);
			}
			_identifier.set_comments(_idCommentList);
		}
	}

	/**
	 * Checks the scope of an identifier and filters out the comment based on
	 * the scope of identifier.
	 * 
	 * @param comment
	 */
	private void FilterComment(Comments comment) {
		if (_identifier == null || _identifier.get_scope() == null)
			return;
		switch (_identifier.get_scope()) {
		/*
		 * case GLOBAL: globalScopeComment(comment); break;
		 */
		case PARENTCLASS:
			parentClassScopeComment(comment);
			break;
		case PACKAGE:
			packageScopeComment(comment);
			break;
		case CLASS:
			classScopeComment(comment);
			break;
		case METHOD:
			methodScopeComment(comment);
			break;
		default:
			break;
		}
	}

	/*
	 * private void globalScopeComment(Comments comment) {
	 * if(_idCommentList!=null) { for(Comments c : _idCommentList) {
	 * if((c.get_comment()).equals(comment.get_comment()) &&
	 * c.get_className()!=null && comment.get_className() !=null &&
	 * (c.get_className()).equals(comment.get_className())) return; } }
	 * _idCommentList.add(comment); }
	 */

	/**
	 * Checks if the parent class name of both identifier and comment is same,
	 * add comment in scope of an identifier. This case usually occurs when one
	 * parent class contains inner class or another private class.
	 * 
	 * @param comment
	 */
	private void parentClassScopeComment(Comments comment) {
		if (_identifier != null && comment.get_parentClassName() != null && _identifier.get_parentClassName() != null
				&& _identifier.get_parentClassName().equalsIgnoreCase(comment.get_parentClassName())) {
			_idCommentList.add(comment);
		}
	}

	/**
	 * Checks if the class name of the comment is same as the class name of the
	 * identifier. Here, if the identifier is of type class, we use the
	 * identifier name as class name. This value is already set while resolving
	 * identifier properties in the IdentifierPropertyResolver class.
	 * 
	 * @param comment
	 */
	private void classScopeComment(Comments comment) {
		if (_identifier != null && comment.get_className() != null && _identifier.get_className() != null
				&& _identifier.get_className().equalsIgnoreCase(comment.get_className())) {
			_idCommentList.add(comment);
		}
	}

	/**
	 * When the scope of an identifier is package level, we check if the comment
	 * package name is same as the identifier package name.
	 * 
	 * @param comment
	 */
	private void packageScopeComment(Comments comment) {
		if (_identifier != null && comment.get_packageName() != null && _identifier.get_packageName() != null
				&& _identifier.get_packageName().equalsIgnoreCase(comment.get_packageName())) {
			_idCommentList.add(comment);
		}
	}

	/**
	 * When an identifier's scope is method, we compare the method name as well
	 * as class names of both identifier and comment. Here we process the line
	 * comments and other comments separately as line comments does not carry
	 * the information about their class node. In that case, we compare if the
	 * compilation unit is same.
	 * 
	 * @param comment
	 */
	private void methodScopeComment(Comments comment) {
		if (comment.get_type() != CommentType.LINE) {
			if (_identifier != null && comment.get_className() != null && comment.get_methodName() != null
					&& _identifier.get_className() != null && _identifier.get_methodName() != null
					&& _identifier.get_className().equalsIgnoreCase(comment.get_className())
					&& _identifier.get_methodName().equalsIgnoreCase(comment.get_methodName())) {
				_idCommentList.add(comment);
			}
		} else {
			if (_identifier != null && comment.get_cuName() != null && _identifier.get_cuName() != null
					&& comment.get_cuName().equals(_identifier.get_cuName())) {
				_idCommentList.add(comment);
			}
		}
	}

	/**
	 * Publishes the results of fragility detection. Creates marker for each
	 * fragile comment linked to the identifier
	 * 
	 * @param markerIds
	 *            contains the previously existing marker ids.
	 * @param identifier
	 *            contain all information about identifier including the list of
	 *            fragile comments associated to it.
	 */
	public void PublishResults(List<IMarker> markerIds, Identifier identifier) {
		this._identifier = identifier;
		for (Comments comment : identifier.get_comments()) {
			for (Match match : comment.get_matches()) {
				this._currentMatch = match;
				// _lineNumber=comment.get_lineNumber();
				_lineNumber = comment.get_lineNumber() + (match.get_matchPosition() - 1);
				_resource = comment.get_cUnit().getResource();
				String message = "Fragile comment detected on changing identifier-" + identifier.get_name();
				createMarker(_resource, message, _lineNumber, markerIds, comment.get_comment().length());
			}
		}
	}

	private void createMarker(IResource aResource, String aMessage, int line, List<IMarker> markers, int length) {
		try {
			final String KEY = "key";
			final String VALUE = "value";
			final String LENGTH = "length";
			boolean isDuplicate = false;
			String match = _currentMatch.get_matchString().substring(0,
					_currentMatch.get_matchString().indexOf("#") - 1);
			IMarker currentMarker = aResource.createMarker(IMarker.PROBLEM);
			currentMarker.setAttribute(IMarker.PRIORITY, IMarker.PRIORITY_HIGH);
			currentMarker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_WARNING);
			currentMarker.setAttribute(IMarker.MESSAGE, aMessage);
			currentMarker.setAttribute(IMarker.LINE_NUMBER, line);
			currentMarker.setAttribute(IMarker.TRANSIENT, true);
			currentMarker.setAttribute(KEY, match);
			currentMarker.setAttribute(VALUE, AstBuilder._newName);
			currentMarker.setAttribute(LENGTH, length);
			currentMarker.setAttribute("original", _identifier.get_name());

			if (markers.size() == 0)
				markers.add(currentMarker);
			else {
				for (IMarker marker : markers) {
					if (marker.exists()) {
						String key = (String) marker.getAttribute(KEY);
						String value = (String) marker.getAttribute(VALUE);
						int lineNo = (int) marker.getAttribute(IMarker.LINE_NUMBER);
						// long markerid=Long.parseLong(id);
						if (key.equals(match) && value.equals(AstBuilder._newName) && lineNo == line) {
							// double check if marker already existing in the
							// resource file.
							// IMarker
							// existingMarker=aResource.findMarker(markerid);
							// if(existingMarker!=null)
							currentMarker.delete();
							isDuplicate = true;
						}
					}
				}
			}
			if (!isDuplicate)
				markers.add(currentMarker);
		} catch (CoreException ce) {
			ce.printStackTrace();
		}
	}
}
