/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/
package ca.mcgill.cs.stg.fraco.parsers;

import java.util.Optional;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;

import ca.mcgill.cs.stg.fraco.AstBuilder;
import ca.mcgill.cs.stg.fraco.utils.Identifier;
import ca.mcgill.cs.stg.fraco.utils.Scope;
import ca.mcgill.cs.stg.fraco.utils.Type;
import ca.mcgill.cs.stg.fraco.visitors.LocalVariableVisitor;

public class LocalsParser extends ParseAstNodes {

	/**
	 * Stores the name of the local declaration parsed.
	 */
	private Optional<String> localName = Optional.ofNullable(null);

	/**
	 * Stores the method name of the local declaration parsed.
	 */
	private String methodName;

	/**
	 * Class Name of the parent of the local under parsing.
	 */
	private String parentName;

	/**
	 * the local declaration under parsing
	 */
	private SingleVariableDeclaration localSVD = null;

	/**
	 * the local declaration under parsing
	 */
	private VariableDeclarationStatement localVDS = null;

	/**
	 * Index of the leading method comment in comment list used to store all the
	 * comments in a compilation unit.
	 */
	private int leadingCommentIndex;

	private boolean localFound;

	/**
	 * Parse all local type AST nodes in compilation Unit.
	 */
	public void parseLocalVariables() {
		LocalVariableVisitor visitor = new LocalVariableVisitor();
		AstBuilder._currentCompilationUnit.accept(visitor);

		for (SingleVariableDeclaration var : visitor.getVariableDeclarations()) {
			localName = Optional.of(var.getName().toString());
			this.localSVD = var;
			this.localVDS = null;
			checkPreConditions();
			if (localFound) {
				break;
			}
		}

		for (VariableDeclarationStatement variable : visitor.getVariableStatements()) {
			String variableName = ((VariableDeclarationFragment) variable.fragments().get(0)).getName().getIdentifier();
			localName = Optional.of(variableName);
			this.localSVD = null;
			this.localVDS = variable;
			checkPreConditions();
			if (localFound) {
				break;
			}
		}
	}

	private void checkPreConditions() {
		if (AstBuilder._renamedElement != null) {
			getRenamedLocalInfo();
		}
	}

	/**
	 * Get the properties required to parse the local. Properties- local name (
	 * need to check it against the renamed local to ensure the right local
	 * parsing) Index of the leading comment for the parsed local. Parent class
	 * name of the local begin parsed.
	 */

	private void getLocalProperties() {
		if (localSVD != null) {
			this.leadingCommentIndex = AstBuilder._currentCompilationUnit.firstLeadingCommentIndex(localSVD);
			this.methodName = getLocalVariableMethodName(localSVD.getParent());
			this.parentName = ParseAstNodes.getClassName(localSVD.getParent(), true);
		} else {
			this.leadingCommentIndex = AstBuilder._currentCompilationUnit.firstLeadingCommentIndex(localVDS);
			this.methodName = getLocalVariableMethodName(localVDS.getParent());
			this.parentName = ParseAstNodes.getClassName(localVDS.getParent(), true);
		}

	}

	private void setRenamedLocalProperties() {
		AstBuilder._identifier.set_type(Type.LOCAL);
		AstBuilder._identifier.set_parentClassName(parentName);
		AstBuilder._identifier.set_methodName(methodName);
		//_identifier.set_cuName(_currentCompilationUnit.getJavaElement().getElementName());
		ParseAstNodes.traverseLeadingComments(leadingCommentIndex);
		if (localSVD != null) {
			int classModifier = ParseAstNodes.getClassModifier(parentName, localSVD.getParent());
			int mod = localSVD.getModifiers();
			AstBuilder._identifier.set_modifier(ParseAstNodes.mapModifier(mod));
			AstBuilder._identifier.set_parentModifier(ParseAstNodes.mapModifier(classModifier));

		} else {
			int classModifier = ParseAstNodes.getClassModifier(parentName, localVDS.getParent());
			int mod = localVDS.getModifiers();
			AstBuilder._identifier.set_modifier(ParseAstNodes.mapModifier(mod));
			AstBuilder._identifier.set_parentModifier(ParseAstNodes.mapModifier(classModifier));
		}
	}

	private void getRenamedLocalInfo() {
		getLocalProperties();
		if (localName.equals(AstBuilder._renamedElement.getElementName())) {
			setRenamedLocalProperties();
			localFound = true;
		} else {
			MethodParser methodParser = new MethodParser();
			methodParser.parseMethods(localName);
		}
	}

	private String getLocalVariableMethodName(ASTNode node) {
		if (node == null)
			return null;
		if (node instanceof MethodDeclaration) {
			return ((MethodDeclaration) node).getName().getIdentifier();
		} else
			return getLocalVariableMethodName(node.getParent());

	}
	// ParseStatements(compilationUnit);
}
