/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/
/**
 * 
 */

package ca.mcgill.cs.stg.fraco.parsers;

import java.util.List;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.EnumDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.TypeDeclaration;

import ca.mcgill.cs.stg.fraco.AstBuilder;
import ca.mcgill.cs.stg.fraco.utils.Comments;
import ca.mcgill.cs.stg.fraco.utils.Identifier;
import ca.mcgill.cs.stg.fraco.utils.Modifiers;
import ca.mcgill.cs.stg.fraco.utils.Scope;
import ca.mcgill.cs.stg.fraco.utils.Type;

/**
 * @author inderjot
 *
 */
public class ParseAstNodes {
	/**
	 * Return the class modifier integer representation
	 * 
	 * @param parentName
	 *            parentNode for which we need modifier information
	 * @param node
	 *            the parent AST node
	 * @return
	 */
	public static int getClassModifier(String parentName, ASTNode node) {
		if (node != null) {
			if (node instanceof TypeDeclaration) {
				List<TypeDeclaration> types = getTopLevelClass();
				if (types == null)
					return 0;
				for (TypeDeclaration type : types) {
					if (type.getName().getIdentifier().equals(parentName))
						return type.getModifiers();
				}
			} else if (node instanceof EnumDeclaration
					&& ((EnumDeclaration) node).getName().getIdentifier().equals(parentName))
				return ((EnumDeclaration) node).getModifiers();
		}
		return 0;
	}

	/**
	 * @returns a list of top level class declarations in the current
	 *          compilation unit.
	 */
	@SuppressWarnings("unchecked")
	private static List<TypeDeclaration> getTopLevelClass() {
		if (AstBuilder._currentCompilationUnit.types() != null && AstBuilder._currentCompilationUnit.types().size() > 0
				&& AstBuilder._currentCompilationUnit.types().get(0) instanceof TypeDeclaration) {
			return AstBuilder._currentCompilationUnit.types();
		}
		return null;
	}

	private static String getTypeName(ASTNode node) {
		if (node == null)
			return null;
		if (node instanceof TypeDeclaration) {
			TypeDeclaration parentType = (TypeDeclaration) node;
			List<TypeDeclaration> types = getTopLevelClass();
			if (types != null) {
				for (TypeDeclaration type : types) {
					if (type.getName() != null && type.isPackageMemberTypeDeclaration()
							&& parentType.getName() == type.getName())
						return type.getName().getIdentifier();
					else if (type.isMemberTypeDeclaration() && type.resolveBinding().getDeclaringClass() != null) {
						TypeDeclaration declaringType = (TypeDeclaration) type.resolveBinding().getDeclaringClass();
						if (declaringType.getName() == parentType.getName())
							return declaringType.getName().getIdentifier();
					}
				}
			}
		} else if (node instanceof EnumDeclaration)
			return ((EnumDeclaration) node).getName().getIdentifier();
		return null;
	}

	/**
	 * Gets class name of the comment node
	 * 
	 * @param node
	 *            the comment node for which we need the class name
	 * @return
	 */
	public static String getClassName(ASTNode node, boolean... isComment) {
		String className = getTypeName(node);
		if (className != null)
			return className;
		else if (isComment.length > 0 && isComment[0])
			return AstBuilder._currentCompilationUnit.getJavaElement().getElementName();

		return null;
	}

	/**
	 * traversing the leading comments for any method, field or class based on
	 * the index and return true of a comment is found for that particular
	 * index.
	 * 
	 * @param index
	 *            the index of the leading comment
	 * @return true if comments list contains a comment on the given index
	 */
	public synchronized static boolean traverseLeadingComments(int index) {
		if (index > -1) {
			String c = AstBuilder._commentList.get(index);
			for (Comments comment : AstBuilder._commentsTable) {
				String commentWithoutNewLine = (comment.get_comment().toString());
				if (commentWithoutNewLine.equals(c)) {
					comment.set_isProximityComment(true);
				}
			}
			return true;
		}
		return false;
	}

	public static Modifiers mapModifier(int mod) {
		if (Modifier.isDefault(mod))
			return Modifiers.DEFAULT;
		else if (Modifier.isPrivate(mod))
			return Modifiers.PRIVATE;
		else if (Modifier.isProtected(mod))
			return Modifiers.PROTECTED;
		else if (Modifier.isPublic(mod))
			return Modifiers.PUBLIC;
		return null;
	}

	public static void markAllInnerCommentsAsLocal(Identifier identifier) {
		if (identifier.get_type() == Type.CLASS) {
			for (Comments comment : AstBuilder._commentsTable) {
				if (comment.get_cuName() != null && comment.get_cuName().equals(identifier.get_cuName())) {
					if (identifier.get_leadingComments() != null)
						identifier.get_leadingComments().add(comment);
					if (AstBuilder._renamedElement != null)
						comment.set_isProximityComment(true);
				}
			}
		}
	}
}
