/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/
package ca.mcgill.cs.stg.fraco.parsers;

import org.eclipse.jdt.core.dom.AnnotationTypeDeclaration;
import org.eclipse.jdt.core.dom.EnumDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;

import ca.mcgill.cs.stg.fraco.AstBuilder;
import ca.mcgill.cs.stg.fraco.utils.Identifier;
import ca.mcgill.cs.stg.fraco.utils.Type;
import ca.mcgill.cs.stg.fraco.visitors.ClassVisitor;

public class TypeParser extends ParseAstNodes {

	private TypeDeclaration type;
	private EnumDeclaration enumm;
	private AnnotationTypeDeclaration annotationType;

	private int mod;
	private int classModifier;
	/**
	 * Stores the class name of the type declaration parsed.
	 */
	private String className;

	/**
	 * Class Name of the parent of the class under parsing.
	 */
	private String parentName;

	/**
	 * Index of the leading method comment in comment list used to store all the
	 * comments in a compilation unit.
	 */
	private int leadingCommentIndex;

	private boolean classFound;

	public synchronized void parseTypes() {
		ClassVisitor visitor = new ClassVisitor();
		AstBuilder._currentCompilationUnit.accept(visitor);
		parseInterfaces(visitor);
		for (TypeDeclaration clazz : visitor.getClasses()) {
			this.type = clazz;
			getTypeProperties();
			if (AstBuilder._renamedElement != null) {
				getRenamedTypeInfo();
				if (classFound) {
					break;
				}
			}
		}
		parseEnums(visitor);
	}

	private void parseInterfaces(ClassVisitor visitor) {
		for (AnnotationTypeDeclaration clazz : visitor.getInterfaces()) {
			this.annotationType = clazz;

			getAnnotationTypeProperties();
			if (AstBuilder._renamedElement != null) {
				getRenamedTypeInfo();
				if (classFound) {
					break;
				}
			}
		}
	}

	private void parseEnums(ClassVisitor visitor) {
		for (EnumDeclaration clazz : visitor.getEnums()) {
			this.enumm = clazz;
			getEnumProperties();
			if (AstBuilder._renamedElement != null) {

				getRenamedTypeInfo();
				if (classFound) {
					break;
				}
			}
		}
	}

	private void getRenamedTypeInfo() {
		if (className.equals(AstBuilder._renamedElement.getElementName())) {
			setRenamedTypeProperties();
			classFound = true;
		}
	}

	private void setRenamedTypeProperties() {
		AstBuilder._identifier.set_type(Type.CLASS);
		AstBuilder._identifier.set_modifier(ParseAstNodes.mapModifier(mod));
		AstBuilder._identifier.set_parentModifier(ParseAstNodes.mapModifier(classModifier));
		// _identifier.set_cuName(_currentCompilationUnit.getJavaElement().getElementName());
		ParseAstNodes.traverseLeadingComments(leadingCommentIndex);
	}

	/**
	 * Get the properties required to parse the class. Properties- class name (
	 * need to check it against the renamed class to ensure the right local
	 * parsing) Index of the leading comment for the parsed class. Parent class
	 * name of the class begin parsed.
	 */
	private void getTypeProperties() {
		className = type.getName().getIdentifier();
		leadingCommentIndex = AstBuilder._currentCompilationUnit.firstLeadingCommentIndex(type);
		parentName = ParseAstNodes.getClassName(type.getParent());
		mod = type.getModifiers();
		classModifier = ParseAstNodes.getClassModifier(parentName, type.getParent());
	}

	/**
	 * Get the properties required to parse the class. Properties- class name (
	 * need to check it against the renamed class to ensure the right local
	 * parsing) Index of the leading comment for the parsed class. Parent class
	 * name of the class begin parsed.
	 */
	private void getEnumProperties() {
		className = enumm.getName().getIdentifier();
		leadingCommentIndex = AstBuilder._currentCompilationUnit.firstLeadingCommentIndex(enumm);
		parentName = ParseAstNodes.getClassName(enumm.getParent());
		mod = enumm.getModifiers();
		classModifier = ParseAstNodes.getClassModifier(parentName, enumm.getParent());
	}

	private void getAnnotationTypeProperties() {
		className = annotationType.getName().getIdentifier();
		leadingCommentIndex = AstBuilder._currentCompilationUnit.firstLeadingCommentIndex(annotationType);
		parentName = ParseAstNodes.getClassName(annotationType.getParent());
		mod = annotationType.getModifiers();
		classModifier = ParseAstNodes.getClassModifier(parentName, annotationType.getParent());
	}
}
