/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/
/**
 * 
 */
package ca.mcgill.cs.stg.fraco.rules;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import ca.mcgill.cs.stg.fraco.utils.Identifier;
import ca.mcgill.cs.stg.fraco.utils.Modifiers;
import ca.mcgill.cs.stg.fraco.utils.Scope;
import ca.mcgill.cs.stg.fraco.utils.Type;

/**
 * This class (singleton) implements scoping ca.mcgill.cs.stg.fraco.rules to
 * fins the scope of an identifier.
 * 
 * @author inderjot
 *
 */
public class ScopeRules {

	/**
	 * This list contains scoping ca.mcgill.cs.stg.fraco.rules for the
	 * identifier of type class. The first modifier belongs to the identifier
	 * itself, second modifier represents identifier's parent and third "Scope"
	 * contains the final value of scope given an identifier matches the first
	 * two values. All these values are stored in the utility class
	 * "Identifier".
	 */
	@SuppressWarnings("serial")
	private static final List<Identifier> SCOPE_LIST_CLASS = Collections.unmodifiableList(new ArrayList<Identifier>() {
		{
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PUBLIC, Scope.GLOBAL));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PROTECTED, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PUBLIC, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.DEFAULT, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PROTECTED, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PUBLIC, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PUBLIC, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PUBLIC, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PROTECTED, Scope.PACKAGE));

			add(getIdentifier(Modifiers.PUBLIC, null, Scope.GLOBAL));
			add(getIdentifier(Modifiers.PRIVATE, null, Scope.CLASS));
			add(getIdentifier(Modifiers.PROTECTED, null, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, null, Scope.PACKAGE));

		}
	});

	/**
	 * This list contains scoping ca.mcgill.cs.stg.fraco.rules for the
	 * identifier of type method. The first modifier belongs to the identifier
	 * itself, second modifier represents identifier's parent and third "Scope"
	 * contains the final value of scope given an identifier matches the first
	 * two values. All these values are stored in the utility class
	 * "Identifier".
	 */
	@SuppressWarnings("serial")
	private static final List<Identifier> SCOPE_LIST_METHOD = Collections.unmodifiableList(new ArrayList<Identifier>() {
		{
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PUBLIC, Scope.GLOBAL));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PROTECTED, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PUBLIC, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PROTECTED, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PUBLIC, Scope.CLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.DEFAULT, Scope.CLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PRIVATE, Scope.CLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PROTECTED, Scope.CLASS));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PUBLIC, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PRIVATE, Scope.CLASS));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PROTECTED, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PRIVATE, Scope.CLASS));

		}
	});

	/**
	 * This list contains scoping ca.mcgill.cs.stg.fraco.rules for the
	 * identifier of type fields. The first modifier belongs to the identifier
	 * itself, second modifier represents identifier's parent and third "Scope"
	 * contains the final value of scope given an identifier matches the first
	 * two values. All these values are stored in the utility class
	 * "Identifier".
	 */
	@SuppressWarnings("serial")
	private static final List<Identifier> SCOPE_LIST_FIELDS = Collections.unmodifiableList(new ArrayList<Identifier>() {
		{
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PRIVATE, Scope.CLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.DEFAULT, Scope.CLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PROTECTED, Scope.CLASS));
			add(getIdentifier(Modifiers.PRIVATE, Modifiers.PUBLIC, Scope.CLASS));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PROTECTED, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PROTECTED, Modifiers.PUBLIC, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PROTECTED, Scope.PACKAGE));
			add(getIdentifier(Modifiers.DEFAULT, Modifiers.PUBLIC, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PRIVATE, Scope.PARENTCLASS));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.DEFAULT, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PROTECTED, Scope.PACKAGE));
			add(getIdentifier(Modifiers.PUBLIC, Modifiers.PUBLIC, Scope.GLOBAL));

		}
	});

	private ScopeRules() {
	}

	/**
	 * Create an object of class "Identifier" to contain the values specific to
	 * each scoping rule.
	 * 
	 * @param idMod
	 *            modifier of the identifier under processing
	 * @param parent
	 *            modifier of identifier's parent.
	 * @param scope
	 *            the final value of scope.
	 * @return the object identifier to add into the scope
	 *         ca.mcgill.cs.stg.fraco.rules list.
	 */
	private static Identifier getIdentifier(Modifiers idMod, Modifiers parent, Scope scope) {
		Identifier id = new Identifier();
		id.set_modifier(idMod);
		id.set_parentModifier(parent);
		id.set_scope(scope);
		return id;
	}

	/**
	 * Gets an identifier under processing and finds the scope of that
	 * identifier using scoping ca.mcgill.cs.stg.fraco.rules list based on the
	 * type of identifier. For local variables, the scope is always set as
	 * "method".
	 * 
	 * @param id
	 */
	public synchronized static void ApplyScopingRules(Identifier id) {

		List<Identifier> _scopeList = new ArrayList<Identifier>();
		setIdType(id);
		if (id != null && id.get_type() != null) {
			switch (id.get_type()) {
			case CLASS:
				_scopeList = SCOPE_LIST_CLASS;
				break;
			case METHOD:
				_scopeList = SCOPE_LIST_METHOD;
				break;
			case FIELD:
				_scopeList = SCOPE_LIST_FIELDS;
				break;
			case LOCAL:
				id.set_scope(Scope.METHOD);
				break;
			default:
				break;

			}
			evaluateIdentifierScope(_scopeList, id);
		}

	}

	private static void setIdType(Identifier id) {
		if (id.get_type() == null && id.get_name().equals(id.get_methodName()))
			id.set_type(Type.METHOD);
	}

	/**
	 * This method sets the scope of identifier by comparing the modifiers.
	 * 
	 * @param scopeList
	 *            the scoping ca.mcgill.cs.stg.fraco.rules list selected on the
	 *            basis of type of the identifier.
	 */
	private static void evaluateIdentifierScope(List<Identifier> scopeList, Identifier id) {
		// filter list based on the identifier scope
		List<Identifier> idscopeList = compareModifier(id.get_modifier(), scopeList, false);
		if (idscopeList != null) {
			// filter list based on the parent scope
			List<Identifier> parentScopeList = compareModifier(id.get_parentModifier(), idscopeList, true);
			if (parentScopeList != null && parentScopeList.size() == 1) {
				id.set_scope(parentScopeList.get(0).get_scope());
			}
		}
	}

	/**
	 * This method compares the modifier of identifier/identifier's parent
	 * against first/second value in an scope list item.
	 * 
	 * @param mod
	 * @param scopeList
	 * @param _isParent
	 * @return
	 */
	private static List<Identifier> compareModifier(Modifiers mod, List<Identifier> scopeList, boolean _isParent) {
		if (mod == null) {
			return FilterScopeList(Modifiers.PUBLIC, scopeList, _isParent);
		} else {
			switch (mod) {
			case PRIVATE:
				return FilterScopeList(Modifiers.PRIVATE, scopeList, _isParent);

			case PUBLIC:
				return FilterScopeList(Modifiers.PUBLIC, scopeList, _isParent);

			case PROTECTED:
				return FilterScopeList(Modifiers.PROTECTED, scopeList, _isParent);

			case DEFAULT:
				return FilterScopeList(Modifiers.DEFAULT, scopeList, _isParent);

			}
		}
		return null;
	}

	/**
	 * This method compares the modifier of the identifier under processing with
	 * the scoping rules list entries. And creates a new list containing the
	 * entries based on modifier match.
	 * 
	 * @param mod
	 *            identifier's modifier or identifier's parent modifier
	 * @param scopeList
	 *            Scoping rules list used to extract the scope.
	 * @param _isParent
	 *            ,true, if the modifier passed belongs to the parent of the
	 *            identifier.
	 * @return returns the filtered list of rules.
	 */
	private static List<Identifier> FilterScopeList(Modifiers mod, List<Identifier> scopeList, boolean _isParent) {
		List<Identifier> filteredScopeList = new ArrayList<Identifier>();
		if (scopeList != null) {
			for (Identifier id : scopeList) {
				if (_isParent) {
					if (id.get_parentModifier() == mod)
						filteredScopeList.add(id);
				} else if (id.get_modifier() == mod)
					filteredScopeList.add(id);
			}
		}
		return filteredScopeList;
	}

}
