/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/
package ca.mcgill.cs.stg.fraco.visitors;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.BlockComment;

import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Javadoc;
import org.eclipse.jdt.core.dom.LineComment;
import org.eclipse.jdt.core.dom.TagElement;

/**
 * CommentVisitor is used to traverse every comment node in AST
 * 
 * @author inder
 *
 */
public class CommentVisitor extends ASTVisitor {
	CompilationUnit cu;
	String source;
	List<String> _commentList;
	List<TagElement> _tagsList = new ArrayList<TagElement>();

	/**
	 * @param cu
	 *            the compilation unit belonging to current AST
	 * @param source
	 * @param commentList
	 *            list of comments present in the current compilation unit. This
	 *            list only provides access to Javadoc comments. To get block
	 *            and line comments, we need comment visitor
	 */
	public CommentVisitor(CompilationUnit cu, String source, List<String> commentList) {
		super();
		this.cu = cu;
		this.source = source;
		this._commentList = commentList;
	}

	/*
	 * (non-Javadoc) Traverses the line comment node in ASt and gets the
	 * description of the comment.
	 * 
	 * @see org.eclipse.jdt.core.dom.ASTVisitor#visit(org.eclipse.jdt.core.dom.
	 * LineComment)
	 */
	public boolean visit(LineComment node) {
		int start = node.getStartPosition();
		int end = start + node.getLength();
		String comment = source.substring(start, end);
		_commentList.add(comment);
		// System.out.println(comment);
		return super.visit(node);
	}

	/*
	 * (non-Javadoc) Traverses the block comment node in ASt and gets the
	 * description of the comment.
	 * 
	 * @see org.eclipse.jdt.core.dom.ASTVisitor#visit(org.eclipse.jdt.core.dom.
	 * BlockComment)
	 */
	public boolean visit(BlockComment node) {
		int start = node.getStartPosition();
		int end = start + node.getLength();
		String comment = source.substring(start, end);
		_commentList.add(comment);
		// System.out.println(comment);

		return super.visit(node);
	}

	/*
	 * (non-Javadoc) Traverses the Javadoc comment node in ASt and gets the tags
	 * description of the comment.
	 * 
	 * @see org.eclipse.jdt.core.dom.ASTVisitor#visit(org.eclipse.jdt.core.dom.
	 * Javadoc)
	 */
	@SuppressWarnings("unchecked")
	public boolean visit(Javadoc node) {
		/*
		 * int start = node.getStartPosition(); int end = start +
		 * node.getLength(); String comment = source.substring(start, end);
		 * _commentList.add(comment);
		 */

		_tagsList.addAll(node.tags());
		return super.visit(node);
	}

	/**
	 * @return the list of tags in a Javadoc comment node
	 */
	public List<TagElement> getTags() {
		return _tagsList;
	}

}
